@@ -1,13 +1,13 @@
 use strict;
 use warnings;
 package BenchmarkAnything::Storage::Frontend::Lib;
-# git description: v0.002-5-g1c4b790
+# git description: v0.003-3-ge09e128
 
 BEGIN {
   $BenchmarkAnything::Storage::Frontend::Lib::AUTHORITY = 'cpan:SCHWIGON';
 }
 # ABSTRACT: Basic functions to access a BenchmarkAnything store
-$BenchmarkAnything::Storage::Frontend::Lib::VERSION = '0.003';
+$BenchmarkAnything::Storage::Frontend::Lib::VERSION = '0.004';
 use Scalar::Util 'reftype';
 
 
@@ -180,6 +180,7 @@ sub _read_config
 {
         my ($self) = @_;
 
+        require File::HomeDir;
         require File::Slurp;
         require YAML::Any;
 
@@ -396,6 +397,163 @@ sub getpoint
         }
 }
 
+sub _multi_point_stats
+{
+        my ($self, $values) = @_;
+
+        require PDL::Stats::Basic;
+        require PDL::Ufunc;
+        require PDL::Core;
+
+        my $data = PDL::Core::pdl(@$values);
+        my $avg  = PDL::Stats::Basic::average($data);
+        return {
+                avg         => PDL::Core::sclr($avg),
+                stdv        => PDL::Stats::Basic::stdv($data),
+                min         => PDL::Ufunc::min($data),
+                max         => PDL::Ufunc::max($data),
+                ci_95_lower => $avg - 1.96 * PDL::Stats::Basic::se($data),
+                ci_95_upper => $avg + 1.96 * PDL::Stats::Basic::se($data),
+               };
+}
+
+
+sub transform_chartlines
+{
+        my ($self, $chartlines, $options) = @_;
+
+        my $x_key       = $options->{x_key};
+        my $x_type      = $options->{x_type};
+        my $y_key       = $options->{y_key};
+        my $y_type      = $options->{y_type};
+        my $aggregation = $options->{aggregation};
+        my $verbose     = defined($options->{verbose}) ? $options->{verbose} : $self->{verbose};
+        my $debug       = defined($options->{debug})   ? $options->{debug}   : $self->{debug};
+        my $dropnull    = $options->{dropnull};
+
+        # from all chartlines collect values into buckets for the dimensions we need
+        #
+        # chartline = title
+        # x         = perlconfig_version
+        # y         = VALUE
+        my @titles;
+        my %VALUES;
+ CHARTLINE:
+        foreach my $chartline (@$chartlines)
+        {
+                my $title     = $chartline->{title};
+                my $results   = $chartline->{results};
+                my $NAME      = $results->[0]{NAME};
+
+                # skip typical empty results
+                if (not @$results or (@$results == 1 and not $results->[0]{NAME}))
+                {
+                        say STDERR "benchmarkanything: transform_chartlines: ignore empty chartline '$title'" if $verbose;
+                        next CHARTLINE;
+                }
+                push @titles, $title;
+
+                say STDERR sprintf("* %-20s - %-40s", $title, $NAME) if $verbose;
+                say STDERR "  VALUE_IDs: ".join(",", map {$_->{VALUE_ID}} @$results) if $debug;
+
+        POINT:
+                foreach my $point (@$results)
+                {
+                        my $x = $point->{$x_key};
+                        my $y = $point->{$y_key};
+                        if (not defined $x)
+                        {
+                                require Data::Dumper;
+                                print STDERR "benchmarkanything: transform_chartlines: chartline '$title': ignore data point (missing key '$x_key'): ".Data::Dumper::Dumper($results) if $verbose;
+                                next POINT;
+                        }
+                        push @{$VALUES{$title}{$x}{values}}, $y; # maybe multiple for same X - average them later
+                }
+        }
+
+        # statistical aggregations of multi points
+        foreach my $title (keys %VALUES)
+        {
+                foreach my $x (keys %{$VALUES{$title}})
+                {
+                        my $multi_point_values     = $VALUES{$title}{$x}{values};
+                        $VALUES{$title}{$x}{stats} = $self->_multi_point_stats($multi_point_values);
+                }
+        }
+
+        # find out all available x-values from all chartlines
+        my %all_x;
+        foreach my $title (keys %VALUES)
+        {
+                foreach my $x (keys %{$VALUES{$title}})
+                {
+                        $all_x{$x} = 1;
+                }
+        }
+        my @all_x = keys %all_x;
+        @all_x =
+         $x_type eq 'version'    ? sort {version->parse($a) <=> version->parse($b)} @all_x
+          : $x_type eq 'numeric' ? sort {$a <=> $b} @all_x
+           : $x_type eq 'string' ? sort {$a cmp $b} @all_x
+            : $x_type eq 'date'  ? sort { die "TODO: sort by date" ; $a cmp $b} @all_x
+             : @all_x;
+
+        # drop complete chartlines if it has gaps on versions that the other chartlines provide values
+        my %clean_chartlines;
+        if ($dropnull) {
+                foreach my $title (keys %VALUES) {
+                        my $ok = 1;
+                        foreach my $x (@all_x) {
+                                if (not @{$VALUES{$title}{$x}{values} || []}) {
+                                        say STDERR "skip: $title (missing values for $x)" if $verbose;
+                                        $ok = 0;
+                                }
+                        }
+                        if ($ok) {
+                                $clean_chartlines{$title} = 1;
+                                say STDERR "okay: $title" if $verbose;
+                        }
+                }
+        }
+
+        # intermediate debug output
+        foreach my $title (keys %VALUES)
+        {
+                foreach my $x (keys %{$VALUES{$title}})
+                {
+                        my $count = scalar @{$VALUES{$title}{$x}{values} || []} || 0;
+                        next if not $count;
+                        my $avg   = $VALUES{$title}{$x}{stats}{avg};
+                        my $stdv  = $VALUES{$title}{$x}{stats}{stdv};
+                        my $ci95l = $VALUES{$title}{$x}{stats}{ci_95_lower};
+                        my $ci95u = $VALUES{$title}{$x}{stats}{ci_95_upper};
+                        say STDERR sprintf("  %-20s . %-7s . avg = %7.2f +- %5.2f (%3d points)", $title, $x, $avg, $stdv, $count) if $verbose;
+                }
+        }
+
+        # result data structure, as needed per chart type
+        my @RESULTMATRIX;
+
+        @titles = grep { !$dropnull or $clean_chartlines{$_} } @titles; # dropnull
+
+        for (my $i=0; $i<@all_x; $i++)          # rows
+        {
+                my $x = $all_x[$i];
+                for (my $j=0; $j<@titles; $j++) # columns
+                {
+                        my $title = $titles[$j];
+                        my $value = $VALUES{$title}{$x}{stats}{$aggregation};
+                        # stringify to unbless from PDL, then numify for type-aware JSON
+                        $value    = $value ? (0+sprintf("%6.2f", $value)) : undef;
+                        $RESULTMATRIX[0]    [0]    = $x_key       if $i == 0 && $j == 0;
+                        $RESULTMATRIX[0]    [$j+1] = $title       if $i == 0;
+                        $RESULTMATRIX[$i+1] [0]    = $x           if            $j == 0;
+                        $RESULTMATRIX[$i+1] [$j+1] = $value;
+                }
+        }
+        return \@RESULTMATRIX;
+}
+
 1;
 
 __END__
@@ -605,6 +763,65 @@ wildcard.
 
 Returns a single benchmark point with B<all> its key/value pairs.
 
+=head2 transform_chartlines ($chartlines, $options)
+
+Gets an array of query results, each one from a different query
+against the backend store, and returns a matrix for rendering those
+chartlines, currently suited for the google charts api.
+
+Multiple results for the same data X-axis are aggregated (default:
+avg).
+
+=over 4
+
+=item INPUT:
+
+  [ title: "dpath-T-n64",
+    results: [
+      {N:dpath, V:1000, version:2.0.13},
+      {N:dpath, V:1170, version:2.0.14},
+      {N:dpath,  V:660, version:2.0.15},
+      {N:dpath, V:1030, version:2.0.16}
+    ]
+  ],
+  [ title: "Mem-nT-n64",
+    results: [
+      {N:Mem,    V:400, version:2.0.13},
+      {N:Mem,    V:460, version:2.0.14},
+      {N:Mem,   V:1120, version:2.0.15},
+      {N:Mem,    V:540, version:2.0.16}
+    ]
+  ],
+  [ title: "Fib-T-64",
+    results: [
+      {N:Fib,    V:100, version:2.0.13},
+      {N:Fib,    V:100, version:2.0.14},
+      {N:Fib,    V:100, version:2.0.15},
+      {N:Fib,    V:200, version:2.0.16}
+    ]
+  ]
+
+=item OUTPUT:
+
+   # multiple results for same version would become aggregated (avg), not show here
+  ['version',   'dpath', 'Mem', 'Fib'],
+  ['2.0.13',      1000,   400,   100],
+  ['2.0.14',      1170,   460,   100],
+  ['2.0.15',       660,  1120,   100],
+  ['2.0.16',      1030,   540,   200]
+
+=back
+
+There are assumptions for the transformation:
+
+=over 4
+
+=item * there is only one NAME per chartline resultset
+
+=item * titles are unique
+
+=back
+
 =head1 AUTHOR
 
 Steffen Schwigon <ss5@renormalist.net>
